package com.vk;

import java.util.*;
import java.io.*;
import java.awt.image.BufferedImage;
import javax.imageio.*;
import com.vk.HexWriter;

public class Img2eps {
    
    private File input;
    private File output;
    private boolean indexed;
    private BufferedImage img;
    private Vector sortedColors = new Vector();
    private int repeatMark;
    private int lastPixel = -1;
    private int sequenceLen = 1;
    private HexWriter hw;
    private int resolution;
    private double scale = 1.0;
    private BufferedWriter fw;
    private int n, mod;
    
    /***************************************************************************
     * Creates a new instance of Bmp2eps
     */
    public Img2eps(String input, String output, int mod) throws Exception {
        this.input = new File(input); this.output = new File(output); this.mod = mod;
        doConversion();
    }
    
    /**************************************************************************/
    public void doConversion() throws Exception {
     img = ImageIO.read(input);
     if (img != null){
      int colors = countUniqueColors(); indexed = (colors <= 255);
      if (indexed) { repeatMark = 255; } else { repeatMark = 0x000001; }
      compressContent();
      FileWriter sfw = new FileWriter(output); fw = new BufferedWriter(sfw);
      writeProlog();
      if (indexed){hw.writeAll(fw, 1);} else {hw.writeAll(fw, 3);}
      writeEpilog();
      fw.close(); sfw.close();
     }else{usage();}
    }
    
    /**************************************************************************/
    private int countUniqueColors() {
        Hashtable colors = new Hashtable();
        // examine pixels
        for (int y=0,maxy=img.getHeight(); y<maxy ; y++) {
            for (int x=0,maxx=img.getWidth(); x<maxx ; x++) {
                // add color to table
                Integer clr = new Integer(img.getRGB(x, y) & 0xFFFFFF);
                Integer num = (Integer)colors.get(clr);
                if (num == null)
                    colors.put(clr, new Integer(1));
                else
                    colors.put(clr, new Integer(num.intValue() + 1));
            }
            if (colors.size() > 1024) break;
        }
        int colorsCount = colors.size();
        if (colorsCount <= 255) {
            Set elems = colors.entrySet();
            Iterator iter = elems.iterator();
            while (iter.hasNext()) {
                Map.Entry elem = (Map.Entry)iter.next();
                Integer col = (Integer)elem.getKey();
                //Integer count = (Integer)elem.getValue();
                //System.out.println("  "+Integer.toHexString(col.intValue())+": "+count);
                sortedColors.add(col);
            }
        }
        colors.clear();
        return colorsCount;
    }

    /**************************************************************************/
    private void compressContent() throws Exception {
        hw = new HexWriter(4096);
        
        // examine pixels
        for (int y=0,maxy=img.getHeight(); y<maxy; y++) {
            for (int x=0,maxx=img.getWidth(); x<maxx; x++) {
                // add color to table
                int px = img.getRGB(x, y) & 0xFFFFFF;
                int out;
                if (indexed) {
                    out = sortedColors.indexOf(new Integer(px));
                    if (out < 0) throw new Exception("Internal error!");
                }
                else {
                    out = px;
                    if (out == 0x000001) out = 0;
                }
                putToCache(out);
            }
            endCaching();
        }
    }
    
    /**************************************************************************/
    private void putToCache(int pixel) throws Exception {
        if (lastPixel == -1) {
            lastPixel = pixel;
            return;
        }
        if (pixel != lastPixel) {
            // different pixel than the previous
            if (sequenceLen <= 1) {
                // no sequence captured
                hw.putValue(lastPixel);
            }
            else {
                // some sequence captured
                if (sequenceLen >= 4) {
                    // the sequence is long enough
                    hw.putValue(repeatMark);
                    if (indexed) {
                        hw.putValue((sequenceLen >> 8) & 0xFF);
                        hw.putValue(sequenceLen & 0xFF);
                    }
                    else {
                        hw.putValue(sequenceLen);
                    }
                    hw.putValue(lastPixel);
                }
                else {
                    // the sequence is too short
                    for (int i=0; i<sequenceLen; i++) hw.putValue(lastPixel);
                }
                sequenceLen = 1;
            }
            lastPixel = pixel;
        }
        else {
            // the same pixel as the previous
            sequenceLen++;
        }
    }

    /**************************************************************************/
    private void endCaching() throws Exception {
        if (sequenceLen <= 1) {
            // no sequence is pending
            hw.putValue(lastPixel);
        }
        else {
            // some pending sequence
            if (sequenceLen >= 4) {
                // the sequence is long enough
                hw.putValue(repeatMark);
                if (indexed) {
                    hw.putValue((sequenceLen >> 8) & 0xFF);
                    hw.putValue(sequenceLen & 0xFF);
                }
                else {
                    hw.putValue(sequenceLen);
                }
                hw.putValue(lastPixel);
            }
            else {
                // the sequence is too short
                for (int i=0; i<sequenceLen; i++) hw.putValue(lastPixel);
            }
        }
        sequenceLen = 1;
        lastPixel = -1;

        hw.newLine();
    }
    
    /**************************************************************************/
    private void wr(String str) throws Exception {
        fw.write(str);
        fw.newLine();
    }
    
    /**************************************************************************/
    private String twoDig(int val) {
        String res = Integer.toString(val);
        if (res.length() < 2) res = "0"+res;
        return res;
    }
    
    private String today() {
        Calendar now = Calendar.getInstance();
        return "" + 
            now.get(Calendar.YEAR) + "-" +
            twoDig(now.get(Calendar.MONTH)+1) + "-" +
            twoDig(now.get(Calendar.DAY_OF_MONTH)) + " " +
            twoDig(now.get(Calendar.HOUR_OF_DAY)) + ":" +
            twoDig(now.get(Calendar.MINUTE)) + ":" +
            twoDig(now.get(Calendar.SECOND));
    }
    
    /**************************************************************************/
    private void writeProlog() throws Exception {
     // header
     int bbw = (int)(img.getWidth());
     int bbh = (int)(img.getHeight());
     if(mod == 0){
      wr("%!PS-Adobe-2.0 EPSF-2.0");
      wr("%%Creator: (Img2eps, written by Petr Slechta, modified by Victor Kohn)");
      wr("%%CreationDate: ("+today()+")");
      wr("%%BoundingBox: 0 0 "+bbw+" "+bbh);
     }
     if (indexed) {
      // store palette
      wr("/pall [");
      for (int i=0,maxi=sortedColors.size(); i<maxi; i++) {
       if ((i > 0) && (i % 8 == 0)) fw.newLine();
       Integer col = (Integer)sortedColors.get(i);
       String item = Integer.toHexString(col.intValue());
       while (item.length() < 6) item = "0" + item;
       fw.write("<"+item+">");
      }
      fw.newLine(); wr("] def"); 
     }
     // prolog for image
     wr("/img [");
    }
    
    /**************************************************************************/
    private void writeEpilog() throws Exception {
        int width = img.getWidth();
        int height = img.getHeight();
        // epilog for image
        wr("] def");
        wr("/lineLen "+width+" 3 mul def  /lineNo 0 def");
        wr("/coded 0 def  /codedPos 0 def  /codedMaxPos 0 def");
        if(indexed){wr("/rd {coded codedPos get  /codedPos codedPos 1 add def} bind def");}
        else{wr("/rd {coded codedPos 3 getinterval  /codedPos codedPos 3 add def} bind def");}
        wr("/pixels 0 def  /pixelsPos 0 def");
        if(indexed){wr("/wr{pall exch get /rgbColor exch def  pixels pixelsPos rgbColor putinterval");}
        else{wr("/wr{pixels exch pixelsPos exch putinterval");}
        wr("  /pixelsPos pixelsPos 3 add def} bind def ");
        if (!indexed){wr("/bin2int {/origStr exch def  origStr 0 get 65536 mul  origStr 1 get 256 mul  origStr 2 get  add  add} bind def");}
        wr("/pixels lineLen string def");
        wr("/decodeLine{/pixelsPos 0 def  /coded img lineNo get def  /codedPos 0 def  /codedMaxPos coded length def");
        wr("  {codedPos codedMaxPos eq {/lineNo lineNo 1 add def  pixels exit} if");
        wr("    /what rd def");
        if (indexed){wr("    what 255 eq");}
        else {wr("    what 0 get 0 eq    what 1 get 0 eq    what 2 get 1 eq    and    and");}
        if (indexed){wr("    {1 1 rd 256 mul rd add");}
        else {wr("    {1 1 rd bin2int");}
        wr("      /what rd def  {pop what wr}bind for");
        wr("    }{what wr} ifelse");
        wr("  } bind loop");
        wr("} bind def");

        int bbw = (int)(width);
        int bbh = (int)(height);

        wr("  gsave");
        wr("  "+bbw+" "+bbh+" scale");
        wr("  "+width+" "+height+" 8 ["+width+" 0 0 -"+height+" 0 "+height+"] { decodeLine } false 3 colorimage");
        wr("  grestore");
        if(mod == 0){wr("%%Trailer");}
    }
    
    /**************************************************************************/
    private static void usage() {
        fileFormats();
    }
    
    /**************************************************************************/
    private static void fileFormats() {
        System.out.println("Image file format is not supported!");
        System.out.println("Supported file formats:");
        String[] formats = ImageIO.getReaderFormatNames();
        for (int i=0,maxi=formats.length; i<maxi; i++)
            System.out.println("  "+formats[i]);
        System.out.println();
    }
    
}
